/*
* Name:        rmccmt.c
* Purpose:     Remove comments from C programs.
* Author:      Lars Wirzenius
* Version:     $Id: rmccmt.c,v 1.2 1993/10/11 21:42:32 liw Exp $
*/

#include <stdio.h>
#include <stdlib.h>

enum state { code, begcmt, incmt, endcmt, str, strquote, chr, chrquote };

int main(int argc, char **argv) {
    enum state st;
    int c, keepnl;

    keepnl = (argc > 1 && strcmp(argv[1], "-n") == 0);

    st = code;
    while ((c = getchar()) != EOF) {
        switch (st) {
        case code:
            switch (c) {
            case '/':
                st = begcmt;
                break;
            case '"':
                st = str;
                putchar(c);
                break;
            case '\'':
                st = chr;
                putchar(c);
                break;
            default:
                putchar(c);
                break;
            }
            break;

        case begcmt:
            switch (c) {
            case '/':
                putchar('/');
                break;
            case '*':
                st = incmt;
                break;
            case '"':
                st = str;
                putchar('/');
                putchar(c);
                break;
            case '\'':
                st = chr;
                putchar('/');
                putchar(c);
                break;
            default:
                putchar('/');
                st = code;
                putchar(c);
                break;
            }
            break;

        case incmt:
            if (c == '\n' && keepnl)
                putchar('\n');
            if (c == '*')
                st = endcmt;
            break;

        case endcmt:
            if (c == '\n' && keepnl)
                putchar('\n');
            if (c == '/') {
                st = code;
                putchar(' '); /* replace cmt with space */
            } else if (c != '*')
                st = incmt;
            break;

        case str:
            putchar(c);
            if (c == '\\')
                st = strquote;
            else if (c == '"')
                st = code;
            break;

        case strquote:
            putchar(c);
            st = str;
            break;

        case chr:
            putchar(c);
            if (c == '\\')
                st = chrquote;
            else if (c == '\'')
                st = code;
            break;

        case chrquote:
            putchar(c);
            st = chr;
            break;

        default:
            fprintf(stderr, "rmccmt: internal logic error\n");
            abort();
        }
    }

    if (st != code) {
        fprintf(stderr, "rmccmt: input is malformed, ends inside %s",
                "comment or literal\n");
        exit(1);
    }
    exit(0);
}
