-- empire lib for  structs V2 converted to use rds libs for the claro cairo wrap project ---
-- this version is unreleased alpha copy and all errors introduced are probably mine --
---------------------------------------------------------------------------------
-- enum.e  CreativePortal.ca,2004
-- version 1.0.0.alpha
--
-- mailto:code@creativeportal.ca
-- http://www.creativeportal.ca                          written by Chris Bensler
---------------------------------------------------------------------------------
-- 
---------------------------------------------------------------------------------

sequence stack    stack = {}
atom count      count = 1
atom inc        inc = 1

------------------------------------------------
global function enum_get_count() --= (atom)
------------------------------------------------
-- syntax:        a = enum_get_count()
-- description:
--    Returns the last enum value generated without incrementing it.
------------------------------------------------
  return count
end function
------------------------------------------------

------------------------------------------------
global function enum_get_inc() --= (atom)
------------------------------------------------
-- syntax:        a = enum_get_inc()
-- description:
--    Returns the value that is used to increment the generated enum values.
------------------------------------------------
  return inc
end function
------------------------------------------------

------------------------------------------------
global procedure enum_set(atom start, atom increment)
------------------------------------------------
-- syntax:        enum_set(a1, a2)
-- description:
--    Changes the current enum state. 'start' will be the first enum value generated.
--    Successive enum values will increase (or decrease) by 'increment'
-- example1:
--    sequence s
--    s = {}
--    enum_set(1,2)
--    for i = 1 to 5 do
--      s &= enum()
--    end for
------------------------------------------------
  inc = increment
  count = start-inc
end procedure
------------------------------------------------

------------------------------------------------
global function enum_start(atom start, atom increment) --= (atom)
------------------------------------------------
-- syntax:        a3 = enum_start(a1,a2)
-- description:
--    Changes the current enum state and returns the first enum value. 'start' will
--    be the first enum value generated. Successive enum values will
--    increase (or decrease) by 'increment'.
-- example1:
--    constant
--        C1 = enum_start(1,1)  -- C1 = 1
--       ,C2 = enum()           -- C2 = 2
--       ,C3 = enum()           -- C3 = 3
-- example2:
--    constant
--        C0  = enum_start(0,-2)  -- C0  = 0
--        C_2 = enum()            -- C_2 = -2
--        C_4 = enum()            -- C_4 = -4
-- See enum_set() for other examples.
------------------------------------------------
  count = start
  inc = increment
 return count
end function
------------------------------------------------


------------------------------------------------
global function enum_push(atom start, atom increment)
------------------------------------------------
-- syntax:        a3 = enum_push(a1,a2)
-- description:
--    Similar to enum_start(), but it pushes the existing enum state onto a stack so it can later be popped.
-- example:
--    constant
--        A0 = enum_start(1,1)  -- A0 = 1
--       ,B0 = enum()           -- B0 = 2
--         ,B1 = enum_push(1,1) -- B1 = 1
--         ,B2 = enum()         -- B2 = 2
--         ,B3 = enum_pop()     -- B3 = 3
--       ,C0 = enum()           -- C0 = 3
-- See enum_set() and enum_start() for other examples.
------------------------------------------------
  stack = { {count,inc} } & stack
  return enum_start(start,increment)
end function
------------------------------------------------

------------------------------------------------
global function enum_pop()
------------------------------------------------
-- syntax:        a3 = enum_pop()
-- description:
--    Compliment to enum_push(). Pops an enum state off the stack.
-- See enum_push() for an example.
------------------------------------------------
 atom ret
  ret = count+inc
  count = stack[1][1]
  inc   = stack[1][2]
  stack = stack[2..length(stack)]
  return ret
end function
------------------------------------------------

------------------------------------------------
global function enum() --= (atom)
------------------------------------------------
-- syntax:        a1 = enum()
-- description:
--    Generates the next enum value.
--    See enum_set() or enum_start() for examples.
------------------------------------------------
  count += inc
 return count
end function
------------------------------------------------

------------------------------------------------
global function enumerate(atom start ,atom stop, atom increment) --= (sequence)
------------------------------------------------
-- syntax:        s = enumerate(a1,a2,a3)
-- description:
--    Generates a sequence of enum values. The first element will be 'start'.
--    Each consecutive element will increase (or decrease) by 'increment'
--    until the enum value reaches 'stop' without exceeding it.
--    An empty sequence will be returned if no valid values can be generated.
-- example1:
--    s = enumerate(1,5,2)        -- returns {1,3,5}
-- example2:
--    s = enumerate(10,1,-1)      -- returns {10,9,8,7,6,5,4,3,2,1}
-- example3:
--     s = enumerate(1,5,-1)      -- returns {}
------------------------------------------------
 sequence s
  s = {}
  for i = start to stop by increment do
    s &= i
  end for
  return s
end function
------------------------------------------------
