-- empire lib for  structs V2 converted to use rds libs for the claro cairo wrap project ---
-- this version is unreleased alpha copy and all errors introduced are probably mine --
--~ include std/other/token.e
include std-types.e

-- reverse find --
global function rfind(object x, sequence s)
  for i = length(s) to 1 by -1 do
    if equal(x,s[i]) then return i end if
  end for
  return 0
end function

-- find all --
global function find_all(object x, sequence s)
 sequence r
  r = {}
  for i = 1 to length(s) do
    if equal(x,s[i]) then r &= i end if
  end for  
  return r
end function

-- TOKEN PARSING ROUTINES --

-- returns a list of tokens from s, delimited by c
global function tokenize(string s, char c)
 sequence token_list,retSeq
 integer start
  retSeq={}
  if length(s) then
    if length(s) != c then s &=c end if
    token_list = find_all(c,s)
    start=1
    for i = 1 to length(token_list) do
      retSeq &={s[start..token_list[i]-1]}
      start = token_list[i]+1
    end for
  end if
  return retSeq
end function

-- returns an inline list of tokens (string), delimited by c
global function inline(sequence tokens, object x)
 sequence s
   s=""
   if length(tokens) then
      s &= tokens[1]
      for i = 2 to length(tokens) do
         s &= x&tokens[i]
      end for
   end if
   return s
end function

-- split off the first token, returns {token,remaining}
global function split_token(sequence s,object x)
 integer f
  if not length(s) then return {"",""} end if
  if atom(x) then x = {x} end if
  f = match(x,s)
  if f <= 1 then return {s,""} end if
  return {s[1..f-1],s[f+length(x)..length(s)]}
end function

-- return the first token in s
global function first_token(sequence s, integer c)
 integer f
  f = find(c,s)
  if f < 1 then return s end if
  return s[1..f-1]
end function

---------------------------------------------------------------------------------
-- CHAR TYPE ROUTINES --

global type digit_char(object c)
  return ('0' <= c) and (c <= '9')
end type

global type hex_uchar(object c)
  return ('A' <= c) and (c <= 'F')
end type

global type hex_lchar(object c)
  return ('a' <= c) and (c <= 'f')
end type

global type hex_char(object c)
  return digit_char(c) or hex_uchar(c) or hex_lchar(c)
end type

global type alpha_uchar(object c)
  return ('A' <= c) and (c <= 'Z')
end type

global type alpha_lchar(object c)
  return ('a' <= c) and (c <= 'z')
end type

global type alpha_char(object c)
  return alpha_uchar(c) or alpha_lchar(c)
end type

global type alphanum_char(object c)
  return alpha_char(c) or digit_char(c)
end type

--~ global type whitespace_char(object c)
 --~ object ret
 --~ sequence wspace
  --~ if atom(c) then
    --~ return (find(c,get_whitespace()) != FALSE)
  --~ else
    --~ wspace = get_whitespace()
    --~ ret = c*0
    --~ for i = 1 to length(wspace) do
      --~ ret = ret or (c = wspace[i])
    --~ end for
    --~ return ret
  --~ end if
--~ end type

global function identifier_char(object c)
  return alphanum_char(c) or (c = '_')
end function

global function getline(integer fn)
 object s
  s = gets(fn)
  if atom(s) then return {} end if
  return s
end function




-- Name: filesys.e
include file.e
global function exists(sequence fname)
        if equal(dir(fname), -1) then
                return 0
        else 
                return 1
        end if
end function


global function parse_path(sequence path)
  if length(path) then
    --~ path = replace(path,'/','\\')
    path = tokenize(path,SLASH)
  end if
  return path
end function

global function parse_filename(sequence fname)
  if length(fname) then
    fname = tokenize(fname,'.')
  end if
  if length(fname) > 1 then
    fname = {inline(fname[1..length(fname)-1],'.'),"."&fname[length(fname)]}
  else
    fname &={ {} }
  end if
  return fname
end function

global function parse_drive(sequence path)
 integer f
  f = find(':',path)
  if f then return path[1..f-1]
  else      return ""
  end if
end function

global function parse_fullpath(sequence path)
 sequence drive,fname
 integer start,stop
  path = parse_path(path)
  start = 1
  stop = length(path)
  drive = parse_drive(path[start])
  fname = parse_filename(path[stop])
  if length(drive) then start +=1 end if
  return {drive, inline(path[start..stop-1],SLASH), fname}
end function
