-- empire lib for  structs V2.017 converted to use rds libs for the claro cairo wrap project ---
-- this version is unreleased alpha copy and all errors introduced are probably mine --
---------------------------------------------------------------------------------
-- structs.e [Version] 2.0.2.alpha [/Version]
--
-- mailto:code@creativeportal.ca
-- http://www.creativeportal.ca                          written by Chris Bensler
---------------------------------------------------------------------------------
--w07117a0:00 fatal_error   dealloc alloc peeks, tests in structs_V2.et



include std-binary.e
include std-enum.e
 --~ include debug.e
-- above include std/mem/mem.e
include std-mem.e
without warning

---------------------------------------------------------------------------------

-- In practice, a programmer could pass a normal, external pointer to this lib and it would happily accept it.
-- The magic number is a kludge used to verify that the pointer was allocated by this library and is allowed to be manipulated.
-- It is not failsafe, however passing an external pointer to these library routines without properly
-- casting them would be a mistake of the programmer and not this lib. The magic number is just to help prevent
-- difficult to track memory leaks caused by these 'runaway' pointers.
constant MAGIC_NUMBER = bytes_to_int("EU C")
constant OFFSET_MAGIC = 4
constant OFFSET_TYPE  = 8
constant OFFSET_COUNT = 12

constant MASK_CLASS    = #F0000000 -- classifier
constant MASK_TYPE     = #0F000000 -- atom/modifier masks
constant MASK_SIZE     = #0000000F -- atomic size masks
constant MASK_INDEX    = #00FFFFFF -- index into structure or union table

-- sizes
-- exclusive (could be inclusive?)
global constant BYTE       = #00000001
global constant WORD       = #00000002
global constant DWORD      = #00000004
global constant QWORD      = #00000008

-- atomic modifiers
-- exclusive
global constant SIGNED     = #01000000
global constant UNSIGNED   = #02000000
global constant FLOAT      = #03000000

global constant EINT       = #06000000
global constant EATOM      = #07000000
global constant ESEQ       = #08000000
global constant EOBJ       = #09000000

global constant BITFIELD   = #0A000000
global constant UNION      = #0B000000
global constant STRUCT     = #0C000000

-- classifiers
-- inclusive
global constant ARRAY      = #40000000
global constant POINTER    = #80000000

-- atomic type enumerations
global constant CHAR    = OR({SIGNED,BYTE})
global constant UCHAR   = OR({UNSIGNED,BYTE})
global constant BOOL    = CHAR
global constant SHORT   = OR({SIGNED,WORD})
global constant USHORT  = OR({UNSIGNED,WORD})
global constant INT     = OR({SIGNED,DWORD})
global constant LONG    = INT
global constant UINT    = OR({UNSIGNED,DWORD})
global constant ULONG   = UINT
global constant FLOAT32 = OR({FLOAT,DWORD})
global constant FLOAT64 = OR({FLOAT,QWORD})
global constant DOUBLE  = FLOAT64

global constant LPSZ    = OR({ARRAY,CHAR})
global constant LPINT   = OR({ARRAY,INT})
global constant LPUINT  = OR({ARRAY,UINT})

-- these 4 are not used by structs.e other than to redefine the names for conformity
global constant INTEGER  = OR({EINT,DWORD})
global constant ATOM     = OR({EATOM,DWORD})
global constant SEQUENCE = OR({ESEQ,DWORD})
global constant OBJECT   = OR({EOBJ,DWORD})


sequence bList      bList = {} -- bitfield properties
sequence uList      uList = {} -- union properties
sequence sList      sList = {} -- struct properties
constant sSIZE = enum_start(1,1)
constant sNAME = enum()
constant sTYPE = enum()
constant sOFFS = enum()
constant sBITS = enum()
constant sNUMB = sBITS

without trace

-- an any_pointer is a dword sized value that is passed to one of the functions in this library and is assumed to be a pointer address
type any_pointer(object x)
  if atom(x) and (floor(x) = x) then
    return TRUE
  else
    return FALSE
  end if
end type

-- a type_pointer is one that is allocated by structs.e ( calloc() )
-- externally allocated pointers should fail
type type_pointer(object x)
  if any_pointer(x) and not AND({x,3}) then -- pointers must be at least 4 byte aligned
    if (peek4u(x-OFFSET_MAGIC) != MAGIC_NUMBER) then -- can this be safely put in the type? or does it need to be checked when type_checking is off?
      fatal_error("Externally allocated pointers require a cast")
    end if
    return TRUE
  end if
  return FALSE
end type

type type_id(object x)
 atom t
  if atom(x) and (floor(x) = x) then
    x = AND({x, NOT(OR({ARRAY,POINTER})) })
    if not x then
      return TRUE
    elsif not AND({ x , MASK_CLASS }) then
      x = AND({ x , NOT(MASK_CLASS) })
      t = AND({ x , MASK_TYPE })
      x = AND({ x , NOT(MASK_TYPE) })
      if (t = BITFIELD) then
        return (x <= length(bList))
      elsif (t = UNION) then
        return (x <= length(uList))
      elsif (t = STRUCT) then
        return (x <= length(sList))
      elsif (t = FLOAT) then
        return (x = DWORD) or (x = QWORD)
      else -- if (t = SIGNED) or (t = UNSIGNED) or (t = BITFIELD) then
        return (x = BYTE) or (x = WORD) or (x = DWORD) or (x = QWORD)
      end if
    end if
  end if
  return FALSE
end type


type type_id_or_array(object x)
  if atom(x) then
    return type_id(x)
  elsif (length(x) = 2) and integer(x[2]) then
    return type_id(x[1])
  end if
  return FALSE
end type

-- a type_cast is of the form {typeID, pointer}
type type_cast(object x)
  if sequence(x) and (length(x) = 2) and type_id(x[1]) and any_pointer(x[2]) then
    return TRUE
  else
    return FALSE
  end if
end type

type type_pointer_or_cast(object x)
  if type_cast(x) then
    return TRUE
  elsif type_pointer(x) then
    return TRUE
  else
    return FALSE
  end if
end type

type type_id_or_cast(object x)
  if type_id(x) then
    return TRUE
  elsif type_cast(x) then
    return TRUE
  else
    return FALSE
  end if
end type

type type_members(object x)
  if sequence(x) then
    return TRUE
  else
    return FALSE
  end if
end type

with trace

global function cast(type_id id, any_pointer lp)
  return {id,lp}
end function


integer typeOf_rID

-- return the index of the first specified member in the struct or union
-- it will work for arrays also, but it is redudnant
-- returns 0 if the member is invalid
global function memberOf(type_id_or_cast x, type_members m)
 atom f,t
 sequence s
  if length(m) then
    if atom(m[1]) then return m[1] end if
    if atom(x) then x = cast(x,NULL) end if
    t = AND({x[1],MASK_TYPE})
    if t = BITFIELD then s = bList
    elsif t = UNION then s = uList
    elsif t = STRUCT then s = sList
    else return 0
    end if
    t = AND({x[1],MASK_INDEX})
    f = find(m[1],s[t][sNAME])
    if f then return f end if
  end if
  fatal_error("Invalid member reference: \'"&m[1]&"\'")
end function

global function typeOf(type_pointer_or_cast lp, type_members m)
 atom id,t
 sequence s
  if atom(lp) then lp = cast(peek4u(lp-OFFSET_TYPE),lp) end if
  id = lp[1]
  lp = lp[2]
  if not length(m) then
    return id
  elsif AND({id,POINTER}) then
    fatal_error("Invalid member reference to a pointer type.")
  elsif AND({id,ARRAY}) then
    if sequence(m[1]) then
      fatal_error("Invaid member reference to an array type.\nExpected an index value, not a string identifier.")
    end if
    return AND({id,NOT(MASK_CLASS)})
  else -- bitfield, union or struct
    t = AND({id,MASK_TYPE})
    if t < BITFIELD then
      fatal_error("Invalid member reference to an atomic type.")
    elsif t = UNION then    s = uList
    elsif t = BITFIELD then s = bList
    else                    s = sList
    end if
    t = s[t][sTYPE][memberOf(cast(t,lp),m)]
    return typeOf(cast(t,lp),m[2..length(m)])
  end if
end function
typeOf_rID = routine_id("typeOf")


-- this function is only valid for array pointers that were allocated using calloc()
-- return the length of the array in units
global function lengthOf(type_pointer lp)
 atom id
  id = typeOf(lp,{})
  if not AND({id,ARRAY}) then fatal_error("Attempt to get the length of an atomic type.") end if
  return peek4u(lp-OFFSET_COUNT)
end function

-- if the type includes POINTER, the size is always 4
-- if the type is a struct or a union, the total size will be returned or the size of the specified member
-- if the type is an array, the unit size of the array will be returned
-- other types are atomic and their sizes are encoded directly into the type identifier. eg INT = SIGNED+DWORD
-- THIS ROUTINE WILL ONLY ACCEPT TYPE IDS
global function sizeOf(type_id_or_cast id, type_members m)
 sequence s
 integer t,size
 atom lp,i
  if atom(id) then id = cast(id,NULL) end if
  lp = id[2]
  id = id[1]
  if AND({id,POINTER}) then
    return 4
  elsif AND({id,ARRAY}) then -- get unit size of array type
    return sizeOf(id-AND({id,ARRAY}),{})
  else
    t = AND({id,MASK_TYPE})
    if t < BITFIELD then -- atomic
      return AND({id,MASK_SIZE})
    elsif t = BITFIELD then s = bList
    elsif t = UNION then    s = uList
    else                    s = sList
    end if
    t = AND({id,MASK_INDEX})
    if length(m) then -- get the size of the specified member
      m[1] = memberOf(id,m)
      i = s[t][sTYPE][m[1]]
      size = sizeOf(i,m[2..length(m)])
      if not AND({i,POINTER}) then
        size *= s[t][sNUMB][m[1]]
      end if
      return size
    else -- get the total size of the struct or union
      return s[t][sSIZE]
    end if
  end if
end function

-- THIS ROUTINE WILL ONLY ACCEPT POINTERS THAT WERE CREATED USING CALLOC()
-- to determine the size of an externally allocated pointer, you must use sizeOf() and specify the type of data
-- there is no way to determine the total lenght of an externally allocated array
global function lpsizeOf(type_pointer lp, type_members m)
 atom size,id
  id = typeOf(lp,{})
  size = sizeOf(id,m)
  if AND({id,ARRAY}) then
    id -= AND({id,ARRAY})
    if length(m) then -- get the size of the specified element
      return sizeOf(id ,m[2..length(m)])
    elsif lp then -- get the total size of the array
      return size*lengthOf(lp)
    -- else -- only an array type was specified, just return the unit size
    end if
  end if
  return size
end function

-- if a pointer is cast, this routine cannot verify that the allocated block is large enough to hold the cast type
global function offsetOf(type_pointer_or_cast lp, type_members m)
 atom id, offset, t
 sequence s
  if atom(lp) then lp = cast(typeOf(lp,{}),lp) end if -- create the cast
  id = lp[1]
  lp = lp[2]
  offset = 0
  if AND({id,POINTER}) then
    if length(m) then fatal_error("Invalid member reference to a pointer type.") end if
    return offset
  elsif AND({id,ARRAY}) then -- array
    if length(m) then
      if not atom(m[1]) then fatal_error("Invald member reference to an array type.\nExpected an index value, not a string identifier") end if
      offset = sizeOf(id,{}) * m[1]-1 -- offset to index
      m = m[2..length(m)]
    end if
    return offset + offsetOf(cast(XOR({id,MASK_CLASS}),lp),m) -- add offset to structure member if any
  else
    t = AND({id,MASK_TYPE})
    if t < BITFIELD then -- atomic type
      if length(m) then fatal_error("Invalid member reference to an atomic type.") end if
      return offset
    elsif t = BITFIELD then s = bList
    elsif t = UNION then    s = uList
    else                    s = sList
    end if
    t = AND({id,MASK_INDEX})
    if length(m) then
      m[1] = memberOf(cast(id,lp),m)
      offset = s[t][sOFFS][m[1]]
      offset += offsetOf(cast(s[t][sOFFS][m[1]],lp),m[2..length(m)])
    end if
    return offset
  end if
end function

global function addressOf(type_pointer_or_cast lp, type_members m)
  if atom(lp) then lp = cast(typeOf(lp,{}),lp) end if
  return lp[1] + offsetOf(lp,m)
end function

global function typedef_bitfield(sequence s)
 sequence struct
 atom size
  struct = {
      0  -- total size
     ,{} -- member names
     ,{} -- member types
     ,{} -- member offsets
     ,{} -- member bitfield
  }
  for i = 1 to length(s) do
    if AND({s[i][2],MASK_TYPE}) >= BITFIELD then
      fatal_error("bitfield."&s[i][1]&": Bitfield member types must be atomic.")
    elsif length(s[i]) < 3 then
      fatal_error("bitfield."&s[i][1]&": Must specify a range for all bitfield members.")
    end if
    struct[sNAME] &={ s[i][1] }
    struct[sTYPE] &={ OR({s[i][2],BITFIELD}) }
    struct[sOFFS] &={ 0 }
    struct[sBITS] &={ s[i][3] }
    size = sizeOf(s[i][2],{})
    if size > struct[sSIZE] then struct[sSIZE] = size end if
  end for
  bList &={ struct }
  return BITFIELD + length(bList)
end function

global function typedef_union(sequence s)
 sequence struct,list
 atom size,t
  struct = {
      0  -- total size
     ,{} -- member names
     ,{} -- member types
     ,{} -- member offsets
     ,{} -- member bitfield
  }
  for i = 1 to length(s) do
    if length(s[i]) < 3 then
      s[i] &= 1
    else
      s[i][2] = OR({s[i][2],ARRAY})
    end if
    struct[sNAME] &={ s[i][1] }
    struct[sTYPE] &={ s[i][2] }
    struct[sOFFS] &={ 0 }
    struct[sNUMB] &={ s[i][3] }
    size = sizeOf(s[i][2],{})
    if not AND({s[i][2],POINTER}) then
      size *= s[i][3]
    end if
    if size > struct[sSIZE] then struct[sSIZE] = size end if

    t = AND({s[i][2],MASK_TYPE})
    if (t = BITFIELD) or (t = STRUCT) then
      if t = BITFIELD then
        list = bList
        struct[sTYPE][length(struct[sTYPE])] = UNSIGNED+AND({s[i][2],#F0FFFFFF})
      else
        list = sList
      end if
      t = AND({s[i][2],MASK_INDEX})
      struct[sNAME] &= list[t][sNAME]
      struct[sTYPE] &= list[t][sTYPE]
      struct[sOFFS] &= list[t][sOFFS]
      struct[sNUMB] &= list[t][sNUMB]
    end if
  end for
  uList &={ struct }
  return UNION + length(uList)
end function

global function typedef_struct(sequence s)
 sequence struct,list
 atom t,offset,size
  struct = {
      0  -- total size
     ,{} -- member names
     ,{} -- member types
     ,{} -- member offsets
     ,{} -- member bitfield
  }
  offset = 0
  for i = 1 to length(s) do
    if length(s[i]) < 3 then
      s[i] &= 1
    else
      s[i][2] = OR({s[i][2],ARRAY})
    end if
    struct[sNAME] &={ s[i][1] }
    struct[sTYPE] &={ s[i][2] }
    struct[sOFFS] &={ offset }
    struct[sNUMB] &={ s[i][3] }

    t = AND({s[i][2],MASK_TYPE})
    if (t = BITFIELD) or (t = UNION) then
      if t = BITFIELD then
        list = bList
        struct[sTYPE][length(struct[sTYPE])] = UNSIGNED+AND({s[i][2],#F0FFFFFF})
      else
        list = uList
      end if
      t = AND({s[i][2],MASK_INDEX})
      struct[sNAME] &= list[t][sNAME]
      struct[sTYPE] &= list[t][sTYPE] 
      struct[sOFFS] &= list[t][sOFFS] + offset
      struct[sNUMB] &= list[t][sNUMB]
    end if

    size = sizeOf(s[i][2],{})
    if not AND({s[i][2],POINTER}) then
      size *= s[i][3]
    end if
    struct[sSIZE] += size
    offset = struct[sSIZE]
  end for
  sList &={ struct }
  return STRUCT + length(sList)
end function

global function calloc(type_id_or_array id)
 atom size,lp,offset
  if atom(id) then id = {id,1} end if
  if id[2] = 1 then
    offset = OFFSET_TYPE
  else
    offset = OFFSET_COUNT
    id[1] = OR({id[1],ARRAY})
  end if
  size = offset + sizeOf(id[1],{}) * id[2]
  lp = allocate(size)
  mem_set(lp,0,size)
  lp += offset
  if id[2] > 1 then poke4(lp-OFFSET_COUNT,id[2]) end if
  poke4(lp-OFFSET_TYPE,id[1])
  poke4(lp-OFFSET_MAGIC,MAGIC_NUMBER)
  return lp
end function

global procedure cfree(type_pointer lp)
 atom id
  id = typeOf(lp,{})
  if AND({id,MASK_TYPE}) >= BITFIELD then
    -- deallocate any members that are owned by this library
  end if
  poke4(lp-OFFSET_MAGIC,NULL) -- zero the magic number, so it doesn't accumulate in garbage ram
  if AND({id,ARRAY}) then
    free(lp-OFFSET_COUNT)-- dealloc
  else
    free(lp-OFFSET_TYPE)-- dealloc
  end if
end procedure

global function cpeek_signed(integer size, any_pointer lp)
  if    size = BYTE then return peeks(lp)
  elsif size = WORD then return peek2s(lp)
  else                   return peek4s(lp)
  end if
end function

global function cpeek_unsigned(integer size, any_pointer lp)
  if    size = BYTE then return peek(lp)
  elsif size = WORD then return peek2u(lp)
  else                   return peek4u(lp)
  end if
end function

global procedure cpoke_int(integer size, any_pointer lp, atom v)
  if    size = BYTE then poke(lp,v)
  elsif size = WORD then poke2(lp,v)
  else                   poke4(lp,v)
  end if
end procedure

global function cpeek_float(integer size, any_pointer lp)
  if size = DWORD then return peek_float32(lp)
  else                 return peek_float64(lp)
  end if
end function

global procedure cpoke_float(integer size, any_pointer lp, atom v)
  if size = DWORD then poke_float32(lp,v)
  else                 poke_float64(lp,v)
  end if
end procedure

global function cpeek_bits(integer size, any_pointer lp, object bits)
 sequence rx
  if atom(bits) then bits = {bits,bits} end if
  rx = int_to_bits(cpeek_unsigned(size,lp),size*8)
  return bits_to_int(rx[bits[1]..bits[2]])
end function

global procedure cpoke_bits(integer size, any_pointer lp, object bits, atom v)
 sequence rx,rv
  if atom(bits) then bits = {bits,bits} end if
  rv = int_to_bits(v,size*8)
  rx = int_to_bits(cpeek_unsigned(size,lp),size*8)
  rx[bits[1]..bits[2]] = rv[1..bits[2]-bits[1]+1]
  cpoke_int(size,lp,bits_to_int(rx))
end procedure

global function cpeek_struct(type_id id, any_pointer lp, type_members m)
 atom t,sz
 object x
  t = AND({id,MASK_INDEX})
  if AND({id,POINTER}) then
    return peek4u(lp)
  elsif AND({id,ARRAY}) then
    id = AND({id, NOT(OR({ARRAY,POINTER}))})
    sz = sizeOf(id,{})
    if length(m) and (sequence(m[1]) or m[1]) then
      m[1] = memberOf(cast(id,lp),m)
      return cpeek_struct(id,lp + sz*(m[1]-1), m[2..length(m)])
    else
      if length(m) then m = m[2..length(m)] end if
      x = repeat(0,lengthOf(lp))
      for i = 1 to length(x) do
        x[i] = cpeek_struct(id,lp,m)
        lp += sz
      end for
      return x
    end if
  elsif AND({id,STRUCT}) then
    if length(m) then
      m[1] = memberOf(cast(id,lp),m)
      return cpeek_struct(sList[t][sTYPE][m[1]],lp+sList[t][sOFFS][m[1]],m[2..length(m)])
    else
      x = repeat(0,length(sList[t][sNAME]))
      for i = 1 to length(x) do
        x[i] = cpeek_struct(sList[t][sTYPE][i],lp+sList[t][sOFFS][i],{sList[t][sNAME][i]})
      end for
      return x
    end if
  elsif AND({id,UNION}) then
    if length(m) then
      x = memberOf(cast(id,lp),m)
      if AND({uList[t][sTYPE][x],MASK_TYPE}) != BITFIELD then m = m[2..length(m)] end if
      return cpeek_struct(uList[t][sTYPE][x],lp,m)
    else
      fatal_error("Ambiguous reference to a union.")
    end if
  elsif AND({id,BITFIELD}) then
    if length(m) then
      return cpeek_bits(sizeOf(id,{}),lp,bList[t][sBITS][memberOf(cast(id,lp),m)])
    else
      fatal_error("No bitfield member specified.")
    end if
  elsif AND({id,FLOAT}) then
    return cpeek_float(sizeOf(id,{}),lp)
  elsif AND({id,UNSIGNED}) then
    return cpeek_unsigned(sizeOf(id,{}),lp)
  else -- signed
    return cpeek_signed(sizeOf(id,{}),lp)
  end if
end function

global procedure cpoke_struct(type_id id, any_pointer lp, type_members m, object data)
 atom t,sz
 object x
  t = AND({id,MASK_INDEX})
  if AND({id,POINTER}) then
    poke4(lp,data)
  elsif AND({id,ARRAY}) then
    id = AND({id, NOT(OR({ARRAY,POINTER}))})
    sz = sizeOf(id,{})
    if length(m) and (sequence(m[1]) or m[1]) then
      m[1] = memberOf(cast(id,lp),m)
      cpoke_struct(id,lp + sz*(m[1]-1), m[2..length(m)], data)
    else -- poke all elements
      if length(m) then m = m[2..length(m)] end if
      x = lengthOf(lp)
      if length(data) > x then data = data[1..x] end if
      for i = 1 to x do
        cpoke_struct(id,lp,m,data[i])
        lp += sz
      end for
    end if
  elsif AND({id,STRUCT}) then
    if length(m) then
      m[1] = memberOf(cast(id,lp),m)
      cpoke_struct(sList[t][sTYPE][m[1]],lp+sList[t][sOFFS][m[1]],m[2..length(m)],data)
    else -- poke all members
       if atom(data) then data = { data } end if  --added  above ln #636/647
      if length(data) > length(sList[t][sNAME]) then
        data = data[1..length(sList[t][sNAME])]
      end if
      for i = 1 to length(data) do
        cpoke_struct(sList[t][sTYPE][i],lp+sList[t][sOFFS][i],{sList[t][sNAME][i]},data[i])
      end for
    end if
  elsif AND({id,UNION}) then
    if length(m) then
      x = memberOf(cast(id,lp),m)
      if AND({uList[t][sTYPE][x],MASK_TYPE}) != BITFIELD then m = m[2..length(m)] end if
      cpoke_struct(uList[t][sTYPE][x],lp+uList[t][sOFFS][x],m,data)
    else
      fatal_error("Ambiguous reference to a union.")
    end if
  elsif AND({id,BITFIELD}) then
    if length(m) then
      m[1] = memberOf(cast(id,lp),m)
      m = bList[t][sBITS][m[1]]
      cpoke_bits(sizeOf(id,{}),lp,m,data)
    else
      fatal_error("No bitfield member specified.")
    end if
  elsif AND({id,FLOAT}) then
    cpoke_float(sizeOf(id,{}),lp,data)
  else -- signed or unsigned
    cpoke_int(sizeOf(id,{}),lp,data)
  end if
end procedure

global function cpeek(type_pointer_or_cast lp, type_members m)
  if atom(lp) then lp = cast(typeOf(lp,{}),lp) end if
  return cpeek_struct(lp[1],lp[2],m)
end function

global procedure cpoke(type_pointer_or_cast lp, type_members m, object data)
  if atom(lp) then lp = cast(typeOf(lp,{}),lp) end if
  cpoke_struct(lp[1],lp[2],m,data)
end procedure

--~ you shold not directly use the constants above CHAR
 --~ Eu pointers are 8 byte aligned
--~ structs.e is byte packed
--w07117a7:14:43 there should be a few smaller pieces so they can be tested seperatly
--as it stands, a failure has to be traced through and none of the other unittests can pass
--m = {0} seems to happen too much

